defmodule IndieWeb.MF2 do
  @moduledoc """
  General purpose Microfrmats2 methods for IndieWeb use.
  """
  require Logger
  alias IndieWeb.URL

  @doc "Extracts a value from a map representing a MF2 item."
  @spec get_value!(map(), binary(), any() | list()) :: binary() | list() | map()
  def get_value!(mf2, property, default_value \\ [])

  def get_value!(%{"properties" => properties}, property, default_value) do
    Map.get(properties, property, default_value)
  end

  def get_value!(_, _, default_value), do: default_value

  # TODO: Define concrete type for the base MF2 object of a page.
  @doc "Extracts all of the MF2 items in the provided structure."
  @spec extract_all(map) :: [map]
  def extract_all(%{"items" => items}) do
    items
    |> Enum.map(&extract_all/1)
    |> List.flatten()
  end

  def extract_all(mf2) when is_map(mf2) do
    mf2
    |> Map.get("children", [])
    |> Enum.flat_map(&extract_all/1)
    |> Enum.concat(extract_from_properties(mf2))
    |> List.insert_at(-1, mf2)
    |> List.flatten()
  end

  def extract_all(mf2, format) do
    mf2
    |> extract_all
    |> Enum.filter(fn item ->
      h_types = item |> Map.get("type", []) |> List.wrap()

      Enum.member?(h_types, "h-#{format}") ||
        Enum.member?(h_types, "h-x-#{format}")
    end)
  end

  def extract_from_properties(%{"properties" => properties}) do
    Enum.flat_map(properties, fn
      {prop, items} ->
        Logger.debug("Scanning property for item.",
          property: prop,
          item: inspect(items)
        )

        items
        |> List.wrap()
        |> List.flatten()
        |> Enum.flat_map(fn
          item when is_map(item) ->
            Logger.debug("Parsing the property with the item.",
              property: prop,
              item: inspect(item)
            )

            extract_all(item)

          _ ->
            Logger.debug("Nothing to do with the provided property; skipping",
              property: prop
            )

            []
        end)
    end)
  end

  def extract_from_properties(_properties) do
    []
  end

  @doc "Parses the provided HTML from `mf2_html` with `base_uri` as a base and returns a map of MF2 data."
  def parse(mf2_html, base_uri)

  def parse(mf2_html, base_uri) when is_nil(mf2_html) or is_nil(base_uri) do
    {:error, :invalid_arguments}
  end

  def parse(mf2_html, base_uri) do
    case Microformats2.parse(mf2_html, base_uri, atomize_keys: false, underscore_keys: true) do
      :error ->
        {:error, :parse_failed}

      mf2 ->
        {:ok, mf2}
    end
  end

  @spec parse!(binary, binary) :: %{optional(atom | binary) => any}
  def parse!(mf2_html, base_uri) do
    case parse(mf2_html, base_uri) do
      {:ok, mf2} -> mf2
      {:error, error} -> raise error
    end
  end

  @doc "Obtains the MF2 object that corresponds to this URL."
  @spec get_item_for_url(map(), binary()) :: {:ok, map()} | {:error, :not_found}
  def get_item_for_url(mf2, url) do
    canonicalized_url = IndieWeb.URL.canonalize(url)

    expected_urls = Enum.uniq([url, canonicalized_url])
    all_items = extract_all(mf2)

    Logger.info("Scanning items for URLs..",
      count: Enum.count(all_items),
      expected_urls: expected_urls
    )

    found_item =
      all_items
      |> Enum.uniq()
      |> Enum.reject(&is_nil/1)
      |> Enum.find(
        nil,
        fn item ->
          item_urls = IndieWeb.MF2.get_value!(item, "url") |> List.wrap()
          canonicalized_item_urls = Enum.map(item_urls, &URL.canonalize/1)

          parsed_urls = Enum.uniq(item_urls ++ canonicalized_item_urls)

          Logger.info("Testing URLs...",
            expected_urls: inspect(expected_urls),
            resolved_urls: inspect(parsed_urls)
          )

          result = Enum.any?(expected_urls, &Enum.member?(parsed_urls, &1))
          Logger.debug("Did any of the URLs match?", result: result)
          result
        end
      )

    first_root_item = mf2 |> Map.get("items", [nil]) |> List.first()

    potential_item =
      [found_item, first_root_item, List.first(all_items)]
      |> Enum.reject(&is_nil/1)
      |> List.first()

    case potential_item do
      nil ->
        {:error, :not_found}

      item ->
        {:ok, item}
    end
  end
end
