defmodule IndieWeb.MF2.Remote do
  @moduledoc """
  Remote interactivity with Web resources.

  This module holds logic for handling remote resources and converting them
  into Microformats2-JSON that can be used by the system.
  """
  alias IndieWeb.Cache
  alias IndieWeb.Http
  alias IndieWeb.MF2
  require Logger

  @doc "Fetches the MF2 of a URI."
  def fetch(uri) do
    Logger.info("Fetching MF2 from the provided URI.", uri: uri)

    case Cache.get(key_func(uri)) do
      nil ->
        Logger.info("URL not found in cache; getting it live.",
          uri: uri
        )

        refresh(uri)

      mf2_json when is_binary(mf2_json) ->
        Logger.info("URL found in cache; using that resource.", uri: uri)
        do_normalize(mf2_json)
    end
  end

  @doc "Refreshes the MF2 for the particular URI."
  @spec refresh(binary()) :: {:ok, map()} | {:error, any()}
  def refresh(uri) do
    case fetch_mf2(uri) do
      {:ok, mf2_json} ->
        store(uri, mf2_json)
        do_normalize(mf2_json)

      {:error, _} = error ->
        error
    end
  end

  @doc "Removes the MF2 cached for the provided MF2."
  @spec flush(binary()) :: :ok
  def flush(uri) do
    Cache.delete(key_func(uri))
  end

  @doc "Stores the provided MF2 to cache."
  @spec store(binary(), map()) :: :ok
  def store(url, mf2_json) do
    Cache.set(key_func(url), mf2_json)
  end

  defp do_normalize(json),
    do: Jason.decode(json, keys: :strings, strings: :copy)

  defp do_serialize(mf2), do: Jason.encode(mf2)

  defp key_func(uri),
    do:
      :crypto.hash(:sha256, uri)
      |> Base.encode64(case: :lower, padding: false)
      |> String.replace_prefix("", "mf2:")

  defp fetch_mf2(uri) do
    with(
      {:ok, %Http.Response{code: code, body: body}}
      when (code >= 200 and code < 300) or code == 410 <- Http.get(uri),
      {:ok, mf2} <- MF2.parse(body, uri)
    ) do
      mf2 |> do_serialize
    else
      {:error, error} ->
        {:error, uri: uri, message: :mf2_remote_fetch_failed, reason: error}

      false ->
        {:error, uri: uri, message: :no_mf2_data_found}

      {:ok, %Http.Response{} = resp} ->
        {:error, uri: uri, message: :no_mf2_data_found, response: resp}
    end
  end
end
