defmodule IndieWeb.Micropub do
  @actionable_keys ~w(replace add delete)
  @reserved_mp_param_keys ~w(h type action url content) ++ @actionable_keys
  @action_prefix "mp-"
  @stable_extensions ~w(post-status published visibility syndication)
  @response_properties ~w(like-of like_of bookmark-of bookmark_of in-reply-to in_reply_to repost-of repost_of)

  @doc "Determines if the provided keys are reserved."
  @spec reserved_keyword?(binary()) :: boolean()
  def reserved_keyword?(key) do
    cond do
      Enum.member?(@reserved_mp_param_keys ++ @stable_extensions, key) -> true
      String.starts_with?(key |> to_string, @action_prefix) -> true
      true -> false
    end
  end

  @doc "Gets the sub-map where the keys are known to be actionable for update operations."
  @spec actionable(params :: map()) :: map()
  def actionable(params), do: Map.take(params, @actionable_keys)

  @doc "Extracts properties that are reserved."
  @spec reserved(properties :: map()) :: map()
  def reserved(properties) do
    keys =
      properties
      |> Map.keys()
      |> Enum.filter(&reserved_keyword?/1)

    Map.take(properties, keys)
  end

  @doc "Separate the extra properties from the reserved properties."
  @spec extract_params(properties :: map()) :: {map(), map()}
  def extract_params(properties) do
    reserved_params = properties |> reserved
    params = Map.drop(properties, Map.keys(reserved_params))

    {params, reserved_params}
  end

  @doc "Fetches the Micropub endpoint for the provided site."
  @spec endpoint_for(binary()) :: binary()
  def endpoint_for(url) do
    IndieWeb.LinkRel.find(url, "micropub") |> List.first()
  end

  def is_response_property?(name) do
    Enum.member?(@response_properties, name)
  end
end
